﻿using System;
using System.Diagnostics;
using System.Collections;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using System.Threading.Tasks;
using System.Net;
using System.Net.Sockets;
using System.Net.NetworkInformation;

namespace Cigno.SCS.Controls
{
    public enum SCSDatagramType : byte
    {
        Command = 0x12,
        EnergyReport = 0x13,
        StatusRequest = 0x15,
        Invalid = 0xFF
    }
    public class SCSDatagram
    {
        public byte Source = 0;
        public byte Destination = 0;
        public byte Payload = 0;
        public SCSDatagramType Type = SCSDatagramType.Command;
        public byte Checksum
        {
            get
            {
                return (byte)(Destination ^ Payload ^ ((byte)Type) ^ Source);
            }
        }
        public SCSDatagram()
        {
            Destination = 0;
            Payload = 0;
            Type = SCSDatagramType.Invalid;
            Source = 0;
        }
        public SCSDatagram(byte Destination, byte Payload, SCSDatagramType Type, byte Source)
        {
            this.Destination = Destination;
            this.Payload = Payload;
            this.Type = Type;
            this.Source = Source;
        }
        public SCSDatagram(SCSDatagram source)
        {
            Copy(source);
        }
        public void Copy(SCSDatagram source)
        {
            Destination = source.Destination;
            Payload = source.Payload;
            Type = source.Type;
            Source = source.Source;
        }
        public static bool TryParse(string data, out SCSDatagram frame)
        {
            data = data.Trim();
            string[] tokens = data.Split(default(string[]), StringSplitOptions.RemoveEmptyEntries);
            for (int i = 0; i < tokens.Length; i++)
            {
                switch (tokens[i].ToUpper())
                {
                    case "K": break; //ignore confirmation response
                    case "E": break; //ignore error response
                    case "E9": break; //command acknowledged (coming from gateway)
                    case "A5": break; //command acknowledged (coming from another device on bus)
                    case "A8":
                        if (tokens.Length - i >= 7 && tokens[i + 6] == "A3") //valid datagram
                        {
                            SCSDatagram tmpFrame = new SCSDatagram();
                            byte.TryParse(tokens[i + 1], System.Globalization.NumberStyles.HexNumber, null as IFormatProvider, out tmpFrame.Destination);
                            byte.TryParse(tokens[i + 2], System.Globalization.NumberStyles.HexNumber, null as IFormatProvider, out tmpFrame.Source);
                            byte.TryParse(tokens[i + 3], System.Globalization.NumberStyles.HexNumber, null as IFormatProvider, out byte tmpType);
                            tmpFrame.Type = (SCSDatagramType)tmpType;
                            byte.TryParse(tokens[i + 4], System.Globalization.NumberStyles.HexNumber, null as IFormatProvider, out tmpFrame.Payload);
                            byte.TryParse(tokens[i + 5], System.Globalization.NumberStyles.HexNumber, null as IFormatProvider, out byte chk);
                            if (chk == tmpFrame.Checksum)
                            {
                                frame = tmpFrame;
                                return true;
                            }
                            break;
                        }
                        break;
                    default: break; //ignore token if not part of a frame
                }
            }
            frame = null;
            return false;
        }
        public override string ToString()
        {
            return string.Format("A8{0:X2}{1:X2}{2:X2}{3:X2}{4:X2}A3", Destination, Source, (byte)Type, Payload, Checksum);
        }
        public bool Equals(SCSDatagram Datagram)
        {
            return (this.Source == Datagram.Source && this.Destination == Datagram.Destination && this.Payload == Datagram.Payload);
        }
    }
    public class SCSCommandDatagram : SCSDatagram
    {
        public byte Command
        {
            get { return Payload; }
        }
        public SCSCommandDatagram(byte Destination, byte Opcode, byte Source) : base(Destination, Opcode, SCSDatagramType.Command, Source) { }
        public SCSCommandDatagram(byte Destination, byte Opcode, bool StatusRequest = false) : base(Destination, Opcode, SCSDatagramType.Command, (byte)(Destination & 0xf0))
        {
            if (StatusRequest)
            {
                base.Type = SCSDatagramType.StatusRequest;
                Source = 0;
            }
        }
    }
    public class SCSDatagramList : LinkedList<SCSDatagram>
    {
        public override string ToString()
        {
            string s = string.Empty;
            foreach (SCSDatagram x in this) { s += x.ToString() + ","; }
            return s;
        }
    }
    public enum SCSCommandStatus
    {
        NotSent,
        SentWaitingConfirmation,
        SuccessfulCompletion,
        Expired,
        Aborted
    }
    public class SCSCommand
    {
        public readonly SCSDatagram CommandDatagram = null;

        DateTime _changedTimestamp = DateTime.MinValue;

        public readonly TimeSpan Timeout = new TimeSpan(0, 0, 0, 0, 900);

        DateTime _expireTimestamp = DateTime.MinValue;
        public DateTime ChangeTimestamp { get { return _changedTimestamp; } }
        public DateTime ExpireTimestamp { get { return _expireTimestamp; } }
        public bool IsExpired { get { return Status == SCSCommandStatus.Aborted || Status == SCSCommandStatus.Expired; } }

        SCSCommandStatus _status = SCSCommandStatus.NotSent;
        public SCSCommandStatus Status
        {
            set
            {
                _status = value;
                _changedTimestamp = DateTime.Now;
                if (_expireTimestamp == DateTime.MinValue || value == SCSCommandStatus.SentWaitingConfirmation)
                {
                    _expireTimestamp = DateTime.Now + Timeout;
                }
                Debug.WriteLine("{2:HH.mm.ss,ffff}\tC:{0},S:{1}", this.CommandDatagram.ToString(), this._status, DateTime.Now);
            }
            get
            {
                if (_status==SCSCommandStatus.SentWaitingConfirmation && (SuccessStatusValues==null || SuccessStatusValues.Length==0))
                {
                    _status = SCSCommandStatus.SuccessfulCompletion;
                }
                if (DateTime.Now > _expireTimestamp)
                {
                    _status = SCSCommandStatus.Expired;
                    Debug.WriteLine("{2:HH.mm.ss,ffff}\tC:{0},S:{1}", this.CommandDatagram.ToString(), this._status, DateTime.Now);
                }
                return _status;
            }
        }
        byte _statusValue = 0xff;
        public byte StatusValue { get { return _statusValue; } }
        byte[] SuccessStatusValues = null;
        public SCSCommand(SCSCommandDatagram Command, byte[] SuccessStatusValues)
        {
            this.CommandDatagram = Command;
            this.SuccessStatusValues = SuccessStatusValues;
            this.Status = SCSCommandStatus.NotSent;
        }
        public SCSCommand(SCSCommandDatagram Command)
        {
            this.CommandDatagram = Command;
            this.SuccessStatusValues = null;
            this.Status = SCSCommandStatus.NotSent;
        }
        public void StatusUpdate(SCSDatagram Datagram)
        {
            if (Status <= SCSCommandStatus.SentWaitingConfirmation)
            {
                /*if (CommandDatagram.Equals(Datagram))
                {
                    Status = SCSCommandStatus.SentWaitingConfirmation;
                }*/
                if (Datagram.Source == CommandDatagram.Destination)
                {
                    Debug.WriteLine("{3:HH.mm.ss,ffff}\tC:{0},R:{1},E:{2}", this.CommandDatagram.ToString(), Datagram.ToString(), SuccessStatusValues != null, DateTime.Now);
                    _statusValue = Datagram.Payload;
                    if (SuccessStatusValues != null)
                    {
                        foreach (byte Value in SuccessStatusValues)
                        {
                            if (Datagram.Payload == Value)
                            {
                                Status = SCSCommandStatus.SuccessfulCompletion;
                            }
                        }
                    }
                    else
                    {
                        Status = SCSCommandStatus.SuccessfulCompletion;
                    }
                    //todo: raise event to notify status subscribers
                }
            }
        }
        public void Abort()
        {
            Status = SCSCommandStatus.Aborted;
        }
    }
    public class SCSCommands : System.Collections.Generic.List<SCSCommand> { }
    public enum SCSControlType
    {
        Unknown,
        OnOff,
        Dimmer,
        Motor,
        Broadcast
    }
    public class SCSControl
    {
        public class SCSControlCommand
        {
            public string Name { get; }
            public SCSDatagram CommandDatagram { get; }
            public byte[] SuccessStatusValues { get; }
            public SCSControlCommand(string Name, SCSDatagram CommandDatagram, byte[] SuccessStatusValues)
            {
                this.Name = Name;
                this.CommandDatagram = CommandDatagram;
                this.SuccessStatusValues = SuccessStatusValues;
            }
        }

        public class SCSControlCommands : Dictionary<string, SCSControlCommand>
        {
            public void Add(SCSControlCommand item)
            {
                base.Add(item.Name, item);
            }
        }

        object Running = new object(); //lock target
        private SCSUDPGateway Gateway = null;
        public string Name = null;
        public readonly byte Address = 0x00;
        private SCSControlType _controlType = SCSControlType.Unknown;
        public SCSControlType ControlType { get { return _controlType; } private set { _controlType = value; } }
        private SCSCommand _lastCommand = null;
        public SCSCommand LastCommand { get { return _lastCommand; } }
        private string _lastReadStatus = "Unknown";
        protected SCSControlCommands _commands;
        public SCSControlCommands.KeyCollection Commands { get { return _commands.Keys; } }
        public string DefaultCommand {
            get {
                switch (ControlType) {
                    case SCSControlType.Broadcast: return "OFF";
                    case SCSControlType.Motor: return "STOP";
                    default: return "TOGGLE";
                }
            }
        } 
        public event EventHandler<string> OnStatusChanged;
        public SCSControl(SCSUDPGateway Gateway, string Name, byte Address, SCSControlType ControlType)
        {
            this.Gateway = Gateway;
            this.Name = Name;
            this.Address = Address;
            this.ControlType = ControlType;
            _commands = new SCSControlCommands();
            if (this.ControlType==SCSControlType.Broadcast && this.Address==0 ) { this.Address = 0xb1; }
            Repurpose(ControlType);
            Gateway.OnNewFrameReceived += AnalyzePacketCallback;
        }
        public void Repurpose(SCSControlType ControlType)
        {
            this.ControlType = ControlType;
            _commands.Clear();
            switch (ControlType)
            {
                case SCSControlType.OnOff:
                    _commands.Add(new SCSControlCommand("TOGGLE", null, null));
                    _commands.Add(new SCSControlCommand("ON", new SCSCommandDatagram(Address, 0), new byte[] { 0 }));
                    _commands.Add(new SCSControlCommand("OFF", new SCSCommandDatagram(Address, 1), new byte[] { 1 }));
                    break;
                case SCSControlType.Dimmer:
                    _commands.Add(new SCSControlCommand("TOGGLE", new SCSCommandDatagram(Address, 2), new byte[] { 0, 1, 0x1d, 0x2d, 0x3d, 0x4d, 0x5d, 0x6d, 0x7d, 0x8d, 0x9d }));
                    _commands.Add(new SCSControlCommand("ON", new SCSCommandDatagram(Address, 0), new byte[] { 0, 0x1d, 0x2d, 0x3d, 0x4d, 0x5d, 0x6d, 0x7d, 0x8d, 0x9d }));
                    _commands.Add(new SCSControlCommand("OFF", new SCSCommandDatagram(Address, 1), new byte[] { 1 }));
                    _commands.Add(new SCSControlCommand("+", new SCSCommandDatagram(Address, 3), new byte[] { 0x2d, 0x3d, 0x4d, 0x5d, 0x6d, 0x7d, 0x8d, 0x9d }));
                    _commands.Add(new SCSControlCommand("-", new SCSCommandDatagram(Address, 4), new byte[] { 0x1d, 0x2d, 0x3d, 0x4d, 0x5d, 0x6d, 0x7d, 0x8d }));
                    _commands.Add(new SCSControlCommand("L1", new SCSCommandDatagram(Address, 0x1d), new byte[] { 0x1d }));
                    _commands.Add(new SCSControlCommand("L2", new SCSCommandDatagram(Address, 0x2d), new byte[] { 0x2d }));
                    _commands.Add(new SCSControlCommand("L3", new SCSCommandDatagram(Address, 0x3d), new byte[] { 0x3d }));
                    _commands.Add(new SCSControlCommand("L4", new SCSCommandDatagram(Address, 0x4d), new byte[] { 0x4d }));
                    _commands.Add(new SCSControlCommand("L5", new SCSCommandDatagram(Address, 0x5d), new byte[] { 0x5d }));
                    _commands.Add(new SCSControlCommand("L6", new SCSCommandDatagram(Address, 0x6d), new byte[] { 0x6d }));
                    _commands.Add(new SCSControlCommand("L7", new SCSCommandDatagram(Address, 0x7d), new byte[] { 0x7d }));
                    _commands.Add(new SCSControlCommand("L8", new SCSCommandDatagram(Address, 0x8d), new byte[] { 0x8d }));
                    _commands.Add(new SCSControlCommand("L9", new SCSCommandDatagram(Address, 0x9d), new byte[] { 0x9d }));
                    break;
                case SCSControlType.Motor:
                    _commands.Add(new SCSControlCommand("UP", new SCSCommandDatagram(Address, 0x8), new byte[] { 0x8 }));
                    _commands.Add(new SCSControlCommand("DOWN", new SCSCommandDatagram(Address, 0x9), new byte[] { 0x9 }));
                    _commands.Add(new SCSControlCommand("STOP", new SCSCommandDatagram(Address, 0xa), new byte[] { 0xa }));
                    break;
                case SCSControlType.Broadcast:
                    _commands.Add(new SCSControlCommand("OFF", new SCSCommandDatagram(0xb1, 0x1, 0x0), null));
                    _commands.Add(new SCSControlCommand("ON", new SCSCommandDatagram(0xb1, 0x0, 0x0), null));
                    _commands.Add(new SCSControlCommand("UP", new SCSCommandDatagram(0xb1, 0x8, 0x0), null));
                    _commands.Add(new SCSControlCommand("DOWN", new SCSCommandDatagram(0xb1, 0x9, 0x0), null));
                    _commands.Add(new SCSControlCommand("STOP", new SCSCommandDatagram(0xb1, 0xa, 0x0), null));
                    _commands.Add(new SCSControlCommand("+", new SCSCommandDatagram(Address, 3), new byte[] { 0x2d, 0x3d, 0x4d, 0x5d, 0x6d, 0x7d, 0x8d, 0x9d }));
                    _commands.Add(new SCSControlCommand("-", new SCSCommandDatagram(Address, 4), new byte[] { 0x1d, 0x2d, 0x3d, 0x4d, 0x5d, 0x6d, 0x7d, 0x8d }));
                    _commands.Add(new SCSControlCommand("L1", new SCSCommandDatagram(Address, 0x1d), new byte[] { 0x1d }));
                    _commands.Add(new SCSControlCommand("L2", new SCSCommandDatagram(Address, 0x2d), new byte[] { 0x2d }));
                    _commands.Add(new SCSControlCommand("L3", new SCSCommandDatagram(Address, 0x3d), new byte[] { 0x3d }));
                    _commands.Add(new SCSControlCommand("L4", new SCSCommandDatagram(Address, 0x4d), new byte[] { 0x4d }));
                    _commands.Add(new SCSControlCommand("L5", new SCSCommandDatagram(Address, 0x5d), new byte[] { 0x5d }));
                    _commands.Add(new SCSControlCommand("L6", new SCSCommandDatagram(Address, 0x6d), new byte[] { 0x6d }));
                    _commands.Add(new SCSControlCommand("L7", new SCSCommandDatagram(Address, 0x7d), new byte[] { 0x7d }));
                    _commands.Add(new SCSControlCommand("L8", new SCSCommandDatagram(Address, 0x8d), new byte[] { 0x8d }));
                    _commands.Add(new SCSControlCommand("L9", new SCSCommandDatagram(Address, 0x9d), new byte[] { 0x9d }));
                    _commands.Add(new SCSControlCommand("_REFRESH", new SCSCommandDatagram(Address, 0, true), null));
                    break;
            }
        }
        public string LastReadStatus
        {
            get { return _lastReadStatus; }
            private set
            {
                _lastReadStatus = value;
                OnStatusChanged?.Invoke(this, _lastReadStatus);
            }
        }
        private void AnalyzePacketCallback(object sender, SCSDatagram e)
        {
            if (e.Source != Address || e.Type != SCSDatagramType.Command)
            {
                return;
            }
            else
            {
                LastReadStatus = GetCommandNameFromOpCode(e.Payload);
            }
        }
        public string GetCommandNameFromOpCode(byte OpCode)
        {
            foreach (SCSControlCommand x in _commands.Values)
            {
                if (x.SuccessStatusValues != null && x.SuccessStatusValues.Length == 1 && x.SuccessStatusValues[0] == OpCode)
                {
                    return x.Name;
                }
            }
            return null;
        }
        private void Send(SCSCommand Command, bool Wait = true, bool Retry = true)
        {
            for (int i = 0; i < (Retry ? 3 : 1); i++)
            {
                DateTime StartTimeStamp = DateTime.Now;
                // 1. let outstanding command execute
                while (Gateway.CountQueuedCommandsForDestination(Command.CommandDatagram.Destination) > 0 && DateTime.Now - StartTimeStamp < Command.Timeout)
                {
                    Thread.Sleep(10);
                    //Thread.Yield();
                }
                // 2. expunge expired commands
                Gateway.ForceDequeueCommandsForDestination(Command.CommandDatagram.Destination);
                // 3. send command
                _lastCommand = Command;
                Gateway.EnqueueCommand(Command);
                do
                {
                    Thread.Sleep(10);
                    //Thread.Yield();
                } while ((!Wait && Command.Status == SCSCommandStatus.SentWaitingConfirmation) ||
                        (Wait && Command.Status == SCSCommandStatus.SentWaitingConfirmation));
                // 4. quit early on success
                if (Command.Status == SCSCommandStatus.SuccessfulCompletion || !Retry || !Wait)
                {
                    if (Command.Status == SCSCommandStatus.SuccessfulCompletion)
                    {
                        LastReadStatus = GetCommandNameFromOpCode(Command.StatusValue);
                    }
                    break;
                }
            }
        }
        /*private void RunCommand(SCSCommand cmd)
        {
            _lastCommand = cmd;
            SendCommandWait(cmd);
        }*/
        public byte GetStatus()
        {
            lock (Running)
            {
                SCSCommand cmd = null;
                switch (this.ControlType)
                {
                    case SCSControlType.Dimmer:
                        cmd = new SCSCommand(new SCSCommandDatagram(Address, 0, true), new byte[] { 0x0, 0x01d, 0x2d, 0x3d, 0x4d, 0x5d, 0x6d, 0x7d, 0x8d, 0x9d });
                        break;
                    case SCSControlType.OnOff:
                        cmd = new SCSCommand(new SCSCommandDatagram(Address, 0, true), new byte[] { 0x0, 0x1 });
                        break;
                    case SCSControlType.Motor:
                        cmd = new SCSCommand(new SCSCommandDatagram(Address, 0, true), new byte[] { 0x8, 0x9, 0xa });
                        break;
                    case SCSControlType.Broadcast:
                        cmd = new SCSCommand(new SCSCommandDatagram(0xb1, 0, true), null);
                        break;
                }
                if (ControlType == SCSControlType.Broadcast)
                {
                    Send(cmd, false, false);
                }
                else
                {
                    //RunCommand(cmd);
                    Send(cmd);
                }
                if (_lastCommand != null)
                {
                    Debug.WriteLine("{3:HH.mm.ss.ffff} C:{0} returned S:{1},SV:{2}", _lastCommand.CommandDatagram.ToString(), _lastCommand.Status, _lastCommand.StatusValue, DateTime.Now);
                    return _lastCommand.StatusValue;
                }
                else
                {
                    //Debug.WriteLine("{3:HH.mm.ss.ffff} C:{0} returned S:{1},SV:{2}", _lastCommand.CommandDatagram.ToString(), null, null, DateTime.Now);
                    return 0xff;
                }
            }
        }
        public void SetStatus(string Command)
        {
            SCSCommand cmd;
            lock (Running)
            {
                if (_commands.ContainsKey(Command))
                {
                    if (_commands[Command].CommandDatagram != null)
                    {
                        cmd = new SCSCommand((SCSCommandDatagram)_commands[Command].CommandDatagram, _commands[Command].SuccessStatusValues);
                        //SendCommandWait(cmd);
                        //RunCommand(cmd);
                        Send(cmd);
                    }
                    else
                    {
                        switch (ControlType)
                        {
                            case SCSControlType.OnOff:
                                switch (Command)
                                {
                                    case "TOGGLE":
                                        switch (this.GetStatus())
                                        {
                                            case 1:
                                                cmd = new SCSCommand((SCSCommandDatagram)_commands["ON"].CommandDatagram, _commands["ON"].SuccessStatusValues);
                                                Send(cmd);
                                                //RunCommand(cmd);
                                                break;
                                            case 0:
                                                cmd = new SCSCommand((SCSCommandDatagram)_commands["OFF"].CommandDatagram, _commands["OFF"].SuccessStatusValues);
                                                Send(cmd);
                                                //RunCommand(cmd);
                                                break;
                                        }
                                        break;
                                }
                                break;
                            case SCSControlType.Dimmer:
                                break;
                            case SCSControlType.Motor:
                                break;
                            case SCSControlType.Broadcast:
                                break;
                        }
                    }
                }
            }
        }
    }
    public class SCSControls : Dictionary<string, SCSControl>
    {
        public SCSControl Broadcast { private set; get; }

        public SCSControls() : base()
        {
        }

        public void Add(SCSControl item)
        {
            base.Add(item.Name, item);
            if (item.ControlType == SCSControlType.Broadcast)
                Broadcast = item;
        }

        public bool ContainsControl(byte Address)
        {
            return Item(Address) != null;
        }
        public bool Repurpose(byte Address, SCSControlType ControlType)
        {
            SCSControl x = Item(Address);
            if (x != null)
            {
                x.Repurpose(ControlType);
                return true;
            }
            return false;
        }
        public bool Rename(byte Address, string NewName)
        {
            SCSControl x = Item(Address);
            if (x != null)
            {
                this.Remove(x.Name);
                x.Name = NewName;
                this.Add(x);
                return true;
            }
            else {
                return false;
            }
        }
        public SCSControl Item(byte Address)
        {
            foreach (SCSControl x in this.Values)
            {
                if (x.Address == Address)
                {
                    return x;
                }
            }
            return null;
        }
        private void DiscoveryReceiver(object sender, SCSDatagram e)
        {
            if (e.Destination == 0xb8 && e.Type == SCSDatagramType.Command && !this.ContainsControl(e.Source))
            {
                SCSControlType Type = SCSControlType.Unknown;
                switch (e.Payload)
                {
                    case 0x0:
                    case 0x1:
                        Type = SCSControlType.OnOff;
                        break;
                    case 0x0d:
                    case 0x1d:
                    case 0x2d:
                    case 0x3d:
                    case 0x4d:
                    case 0x5d:
                    case 0x6d:
                    case 0x7d:
                    case 0x8d:
                    case 0x9d:
                        Type = SCSControlType.Dimmer;
                        break;
                    case 0x8:
                    case 0x9:
                    case 0xa:
                        Type = SCSControlType.Motor;
                        break;
                }
                SCSControl x = Item(e.Source);
                if (x != null)
                {
                    if (x.ControlType < Type) { Repurpose(e.Source, Type); }
                }
                else
                {
                    Add(new SCSControl((SCSUDPGateway)sender, e.Source.ToString(), e.Source, Type));
                }
            }
        }
        public void Discover(SCSUDPGateway Gateway, int Timeout = 3000)
        {
            bool ReceiveStatus = Gateway.IsReceiving;
            Gateway.OnNewFrameReceived += DiscoveryReceiver;
            if (!ReceiveStatus) { Gateway.StartReceiveLoop(); }
            Gateway.Send(new SCSDatagram(0xb1, 0, SCSDatagramType.StatusRequest, 0));
            Thread.Sleep(Timeout);
            Gateway.OnNewFrameReceived -= DiscoveryReceiver;
            if (!ReceiveStatus) { Gateway.StopReceiveLoop(); }
        }
    }
    public class SCSUDPGateway
    {
        public class SCSLog : Queue<SCSLog.SCSLogEntry>
        {
            public class SCSLogEntry
            {
                public readonly DateTime TimeStamp = DateTime.MinValue;
                public readonly string RawData = string.Empty;
                public readonly SCSDatagramList Datagrams = null;
                public SCSLogEntry(DateTime TimeStamp, string RawData, SCSDatagramList Datagrams)
                {
                    this.TimeStamp = TimeStamp;
                    this.RawData = RawData;
                    this.Datagrams = Datagrams;
                }
                public override string ToString()
                {
                    if (Datagrams != null)
                    {
                        return string.Format("{0:HH.mm.ss,ffff} {1,-40}\t[{2,-32}]", TimeStamp, RawData, Datagrams.ToString().Trim().Replace(" ", ","));
                    }
                    else
                    {
                        return string.Format("{0:HH.mm.ss,ffff} {1}", TimeStamp, RawData);
                    }
                }
            }
            public class SCSLogOnDataArgs : EventArgs
            {
                public SCSLogEntry LogEntry { get; set; }
                public SCSLogOnDataArgs(SCSLogEntry item)
                {
                    LogEntry = item;
                }
                public override string ToString()
                {
                    return LogEntry.ToString();
                }
            }

            public bool Enabled = false;
            public SCSLog(int Size) : base(Size)
            {
            }
            public event EventHandler<SCSLogOnDataArgs> OnDataReceived;
            public new void Enqueue(SCSLogEntry item)
            {
                if (Enabled)
                {
                    base.Enqueue(item);
                }
                OnDataReceived?.Invoke(this, new SCSLogOnDataArgs(item));
            }
        }
        class SCSGatewayBuffer
        {
            static readonly string LogDataPattern = @"^SCS\[(?<pos>\d)\]\:((\ (?<byte>[0-9,A-F,a-f]{2}))+)";
            byte GatewayBufferSize = 10;
            byte ReorderingFlexibility = 3;
            TimeSpan ExpirationTimeout = new TimeSpan(0, 0, 0, 0, 500);

            enum ParseStatus
            {
                Failed,
                NotStarted,
                Start,
                Destination,
                Source,
                Type,
                OpCode,
                Checksum,
                Done
            }
            class SCSGatewayBufferEntry
            {
                public readonly DateTime TimeStamp = DateTime.MinValue;
                public readonly sbyte BufferPosition = -1;
                public readonly sbyte ByteOrder = -1;
                public readonly byte Data = byte.MaxValue;
                public SCSGatewayBufferEntry(DateTime TimeStamp, sbyte BufferPosition, sbyte ByteOrder, byte Data)
                {
                    this.TimeStamp = TimeStamp;
                    this.BufferPosition = BufferPosition;
                    this.ByteOrder = ByteOrder;
                    this.Data = Data;
                }
                public override string ToString()
                {
                    return string.Format("T:{0},BP:{1:00},BO:{2:00},D:{3:x2}", TimeStamp, BufferPosition, ByteOrder, Data);
                }
            }
            public string Dump()
            {
                string s = string.Empty;
                foreach (SCSGatewayBufferEntry x in Buffer)
                {
                    s = s + x.ToString() + Separators;
                }
                return s;
            }
            class SCSGatewayBufferEntries : LinkedList<SCSGatewayBufferEntry> { }
            SCSGatewayBufferEntries Buffer = new SCSGatewayBufferEntries();
            public SCSDatagramList ProcessReceivedData(DateTime TimeStamp, string RawData)
            {
                SCSDatagramList Datagrams = null;
                MatchCollection Matches = Regex.Matches(RawData, LogDataPattern);
                if (Matches.Count == 1)
                {
                    sbyte Position = sbyte.Parse(Matches[0].Groups["pos"].Value);

                    // 1. find where to insert
                    LinkedListNode<SCSGatewayBufferEntry> item = Buffer.Last;
                    bool found = false;
                    int i = ReorderingFlexibility;
                    while (item != null && i > 0)
                    {
                        if ((item.Value.BufferPosition + 1) % GatewayBufferSize == Position)
                        {
                            found = true;
                            break;
                        }
                        else
                        {
                            do
                            {
                                item = item.Previous;
                            } while (item != null && item.Value.BufferPosition == item.Next.Value.BufferPosition);
                        }
                        i--;
                    }

                    // 2. insert
                    if (!found)
                    {
                        item = Buffer.Last;
                    }
                    for (i = 0; i < Matches[0].Groups["byte"].Captures.Count; i++)
                    {
                        if (item != null)
                        {
                            item = Buffer.AddAfter(item, new SCSGatewayBufferEntry(TimeStamp, Position, (sbyte)i, byte.Parse(Matches[0].Groups["byte"].Captures[i].Value, System.Globalization.NumberStyles.HexNumber)));
                        }
                        else
                        {
                            item = Buffer.AddFirst(new SCSGatewayBufferEntry(TimeStamp, Position, (sbyte)i, byte.Parse(Matches[0].Groups["byte"].Captures[i].Value, System.Globalization.NumberStyles.HexNumber)));
                        }
                    }

                    // 3. skip and remove old frames
                    while (Buffer.Count > 0 && TimeStamp - Buffer.First.Value.TimeStamp > ExpirationTimeout)
                    {
                        Buffer.RemoveFirst();
                    }

                    // 4. try parse datagrams
                    Datagrams = new SCSDatagramList();
                    item = Buffer.First;
                    ParseStatus ParseStatus = ParseStatus.NotStarted;
                    SCSDatagram tmpFrame = new SCSDatagram();
                    while (item != null)
                    {
                        switch (ParseStatus)
                        {
                            case ParseStatus.Failed:
                                ParseStatus = ParseStatus.NotStarted;
                                break;
                            case ParseStatus.NotStarted:
                                if (item.Value.Data == 0xa8)
                                {
                                    ParseStatus = ParseStatus.Start;
                                }
                                break;
                            case ParseStatus.Checksum:
                                if (item.Value.Data == 0xa3)
                                {
                                    ParseStatus++;
                                }
                                else
                                {
                                    ParseStatus = ParseStatus.Failed;
                                    //backtrack 4 bytes and restart searching
                                    for (i = 0; i < 4 && item.Previous != null; i++) { item = item.Previous; }
                                }
                                break;
                            case ParseStatus.Start:
                                ParseStatus++;
                                tmpFrame.Destination = item.Value.Data;
                                break;
                            case ParseStatus.Source:
                                ParseStatus++;
                                tmpFrame.Type = (SCSDatagramType)item.Value.Data;
                                break;
                            case ParseStatus.Destination:
                                ParseStatus++;
                                tmpFrame.Source = item.Value.Data;
                                break;
                            case ParseStatus.Type:
                                ParseStatus++;
                                tmpFrame.Payload = item.Value.Data;
                                break;
                            case ParseStatus.OpCode:
                                if (tmpFrame.Checksum != item.Value.Data)
                                {
                                    ParseStatus = ParseStatus.Failed;
                                    //backtrack 4 bytes and restart searching
                                    for (i = 0; i < 4 && item.Previous != null; i++) { item = item.Previous; }
                                }
                                else
                                {
                                    ParseStatus++;
                                }
                                break;
                        }

                        if (ParseStatus == ParseStatus.Done)
                        {
                            Datagrams.AddLast(new SCSDatagram(tmpFrame));
                            ParseStatus = ParseStatus.NotStarted;
                            for (LinkedListNode<SCSGatewayBufferEntry> x = Buffer.First; Buffer.Count > 0 && x.Value != item.Value; Buffer.RemoveFirst()) ;
                        }

                        if (item.Next != null &&
                            ((item.Next.Value.BufferPosition == item.Value.BufferPosition && item.Next.Value.ByteOrder == item.Value.ByteOrder + 1) ||
                            (item.Next.Value.BufferPosition == ((item.Value.BufferPosition + 1) % GatewayBufferSize) && item.Next.Value.ByteOrder == 0)))
                        {
                            item = item.Next;
                        }
                        else
                        {
                            break;
                        }
                    }
                }
                return Datagrams;
            }
            public SCSGatewayBuffer() { }
            public SCSGatewayBuffer(byte GatewayBufferSize = 10, byte ReorderingFlexibility = 3, int ExpirationTimeout = 500)
            {
                this.GatewayBufferSize = GatewayBufferSize;
                this.ReorderingFlexibility = ReorderingFlexibility;
                this.ExpirationTimeout = new TimeSpan(0, 0, 0, 0, ExpirationTimeout);
            }
        }
        public static void AutoDiscover(List<IPEndPoint> EndPoints, int Port = 52056, int Timeout = 1500)
        {
            EndPoints.Add(GetGatewayEndPoint(null, Port, Timeout));
        }

        public static bool Probe(string gatewayAddress, int Port = 52056, int Timeout = 900)
        {
            return GetGatewayEndPoint(gatewayAddress, Port, Timeout)!=null;
        }

        public static IPEndPoint GetGatewayEndPoint(string GatewayAddress, int Port = 52056, int Timeout=1500)
        {
            IPAddress TargetAddress = null;
            if (!IPAddress.TryParse(GatewayAddress, out TargetAddress))
            {
                try
                {
                    IPHostEntry HostEntry = Dns.GetHostEntry(GatewayAddress);
                    if (HostEntry != null && HostEntry.AddressList.Length > 0)
                    {
                        TargetAddress = HostEntry.AddressList[0];
                    }
                }
                catch (Exception e)
                {
                    NetworkInterface[] NICs = NetworkInterface.GetAllNetworkInterfaces();
                    foreach (NetworkInterface Adapter in NICs)
                    {
                        if (Adapter.Supports(NetworkInterfaceComponent.IPv4))
                        {
                            IPInterfaceProperties AdapterProperties = Adapter.GetIPProperties();
                            IPv4InterfaceProperties IFProperties = AdapterProperties.GetIPv4Properties();
                            if (IFProperties != null && IFProperties.Index != 0 && !IFProperties.IsAutomaticPrivateAddressingActive)
                            {
                                foreach (UnicastIPAddressInformation x in AdapterProperties.UnicastAddresses)
                                {
                                    if (TargetAddress == null && x.Address.AddressFamily == AddressFamily.InterNetwork)
                                    {
                                        TargetAddress = new IPAddress(x.Address.Address & x.IPv4Mask.Address | (IPAddress.Broadcast.Address & ~x.IPv4Mask.Address));
                                    }
                                }
                            }
                        }
                    }
                }
            }
            IPEndPoint TargetEndPoint = new IPEndPoint(TargetAddress, Port);

            UdpClient Socket = new UdpClient() { EnableBroadcast = true };

            Socket.Send(Encoding.ASCII.GetBytes("@q"), 2, TargetEndPoint);
            DateTime ExpirationTime = DateTime.Now.AddMilliseconds(Timeout);
            IPEndPoint GatewayEndpoint = null;
            //Thread.Sleep(Timeout);
            do
            {
                Thread.Sleep(50);
                if (Socket.Available > 0)
                {
                    string x = Encoding.ASCII.GetString(Socket.Receive(ref GatewayEndpoint));
                    if (Regex.Match(x, "SCS \\d{2}\\.\\d{2}") != null) { break; }
                }
            } while ( DateTime.Now < ExpirationTime );
            Socket.Close();
            return GatewayEndpoint;
        }

        private SCSCommands CommandQueue = new SCSCommands();
        public SCSLog Log = null;
        public IPEndPoint EndPoint;
        UdpClient Socket;
        private object Initialization = new object(); //lock
        Thread ReceiveThread = null;
        static readonly string[] Separators = new string[] { "\r\n" };
        SCSGatewayBuffer Buffer = new SCSGatewayBuffer();
        public SCSUDPGateway(IPEndPoint EndPoint)
        {
            this.EndPoint = EndPoint;
            //Socket = new UdpClient();
            Log = new SCSLog(250);
        }
        ~SCSUDPGateway()
        {
            if (ReceiveThread != null && ReceiveThread.ThreadState != System.Threading.ThreadState.Aborted) { ReceiveThread.Abort(); }
            ReceiveThread = null;
            if (Socket != null) { Socket.Close(); }
            Socket = null;
            EndPoint = null;
            Log = null;
        }
        public void EnqueueCommand(SCSCommand Command)
        {
            CommandQueue.Add(Command);
            Send(Command.CommandDatagram);
            Command.Status = SCSCommandStatus.SentWaitingConfirmation;
        }
        public int CommandsInQueue { get { return CommandQueue.Count; } }
        public int CountQueuedCommandsForDestination(byte Destination)
        {
            int c = 0;
            for (int i = 0; i < CommandQueue.Count; i++) { if (CommandQueue[i].CommandDatagram.Destination == Destination) c++; }
            return c;
        }
        public void ForceDequeueCommandsForDestination(byte Destination)
        {
            for (int i = 0; i < CommandQueue.Count;)
            {
                if (CommandQueue[i].CommandDatagram.Destination == Destination)
                {
                    CommandQueue[i].Abort();
                    try { CommandQueue.RemoveAt(i); } catch { }
                }
                else
                {
                    i++;
                }
            }
        }
        public event EventHandler<SCSDatagram> OnNewFrameReceived;
        public void StartReceiveLoop()
        {
            lock (Initialization)
            {
                if (ReceiveThread != null && ReceiveThread.IsAlive) { return; }
                ReceiveThread = new Thread(DoReceiveLoop);
                Socket = new UdpClient();
                ReceiveThread.Start();
                SendSpecialCommand("@l");
            }
        }
        public bool IsReceiving { get { return ReceiveThread != null && ReceiveThread.IsAlive; } }
        public void StopReceiveLoop()
        {
            lock (Initialization)
            {
                if (ReceiveThread != null)
                {
                    ReceiveThread.Abort();
                    ReceiveThread = null;
                    Socket.Close();
                    Socket = null;
                }
            }
        }
        private void DoReceiveLoop()
        {
            string rawdata = null;
            string[] data = { };
            while (Socket != null)
            {
                if (Socket.Available > 0)
                {
                    rawdata = Encoding.Default.GetString(Socket.Receive(ref this.EndPoint));
                    DateTime TimeStamp = DateTime.Now;
                    if (rawdata != null)
                    {
                        foreach (string line in rawdata.Split(Separators, StringSplitOptions.RemoveEmptyEntries))
                        {
                            SCSDatagramList Datagrams = Buffer.ProcessReceivedData(TimeStamp, line);
                            SCSLog.SCSLogEntry LogEntry = new SCSLog.SCSLogEntry(TimeStamp, line, Datagrams);
                            Log.Enqueue(LogEntry);
                            if (LogEntry.Datagrams != null)
                            {
                                foreach (SCSDatagram Datagram in LogEntry.Datagrams)
                                {
                                    if (CommandQueue != null)
                                    {
                                        for (int i = 0; i < CommandQueue.Count;)
                                        {
                                            if (Datagram.Type == SCSDatagramType.Command && Datagram.Source == CommandQueue[i].CommandDatagram.Destination)
                                            {
                                                if (CommandQueue[i].Status != SCSCommandStatus.SuccessfulCompletion)
                                                {
                                                    CommandQueue[i].StatusUpdate(Datagram);
                                                }
                                                else
                                                {
                                                    Debug.WriteLine("{3:HH.mm.ss.ffff} C:{0} skipped processing of {1} because S:{2}", CommandQueue[i].CommandDatagram.ToString(), Datagram.ToString(), CommandQueue[i].Status, DateTime.Now);
                                                }
                                            }
                                            if (CommandQueue[i].Status >= SCSCommandStatus.SuccessfulCompletion)
                                            {
                                                CommandQueue.RemoveAt(i);
                                            }
                                            else
                                            {
                                                i++;
                                            }
                                        }
                                    }
                                    OnNewFrameReceived?.Invoke(this, Datagram);
                                }
                            }
                            if (CommandQueue != null)
                            {
                                for (int i = 0; i < CommandQueue.Count;)
                                {
                                    if (CommandQueue[i].IsExpired || CommandQueue[i].Status == SCSCommandStatus.SuccessfulCompletion)
                                    {
                                        CommandQueue.RemoveAt(i);
                                    }
                                    else
                                    {
                                        i++;
                                    }
                                }
                            }
                            /*if (Frames != null)
                            {
                                foreach (SCSDatagram x in Frames)
                                {
                                    Console.WriteLine(x.ToString());
                                }
                            }*/
                        }
                    }
                }
                //Thread.Sleep(10);
                Thread.Yield();
            }
        }
        public void Send(SCSDatagram data)
        {
            string CMD = null;
            CMD = "@W7" + data.ToString();
            SendSpecialCommand(CMD);
        }
        public void SendSpecialCommand(string CMD)
        {
            if (Socket == null)
            {
                Socket = new UdpClient();
            }
            Socket.Send(Encoding.ASCII.GetBytes(CMD), CMD.Length, EndPoint);
            if (ReceiveThread == null)
            {
                Socket.Close();
                Socket = null;
            }
            //Thread.Sleep(10);
            //Thread.Yield();
        }
    }
}
